/*
   FlexSensor.h - A Class that provides useful
   functions for using flex-sensors with arduino

   Data will be send via strings and integers.
   When a string is within "$$" like : $$string$$ it's marked as a warning for the user and should be displayed on the GUI.

   A short signal will be indicated by sending a 41, long by 42.

  
*/


#ifndef FLEXSENSOR_h
#define FLEXSENSOR_h



class FlexSensor {
  private:
    //PINS
    int pin;                    //pin to connect sensor
    int onLedPin;               //pin for green status led
    int sleepLedPin;            //pin for red status led

    //STATUS
    int active;                 //can be used to check, if sensor should be collecting data
    int calibrated;             //true if sensor has been calibrated
    int continuousValues;       //when true send difference between average value and momentary value always; deactivates 41/42 sending
    int longClickEnabled;       //when true check for long signals, standard true
    int bothDirections;         //when true return the direction in which the sensor is bend

    //VALUE PROCESSING
    long standardValue;         //average value - reached when sensor is in standard state
    int countSignals;           //signals recognized in a row. used for long signal recognition
    int pSignal = false;        //previous signal, false when no movement detected
    int errorState;             //will be true when a disconnection is assumed


    //SETTINGS
    int tolerance;              //tolerance that determines how harsh the sensor reacts
    int shortThresh = 1;        //time threshold for short-signal definition
    int longThresh = 5;         //time threshold for long-signal definition
    int onLedBrightness = 160;
    int sleepLedBrightness = 200;


    //CONSTANTS
    static const int SHORT = 41;
    static const int LONG = 42;
    static const int SHORTV = 51; //SHORT for negative direction
    static const int LONGV = 52;  //LONG for negative direction



  public:


    //CONSTRUCTOR
    FlexSensor(int);            //construct only with pin number
    FlexSensor(int, int);      //contruct with pin number AND tolerance

    //SETTINGS
    void setLedPins(int onLedPin, int sleepLedPin); //set pin numbers to set lights indicating the state of the arduino
    void enableLongClick(int);    //sets longClickEnabled to true/false
    void enableContinuousValues(int);
    void enableBothDirections(int);

    //ACTION
    void calibrate();           //calls a calibration of the sensor. this will take 1.5 seconds
    void reset();               //resets values, except for tolerance and pin, recalibration is an consequence
    int readData();
    int setActive(int);         //set active state to true/false; if input is not suitable, return 0;


    //VALUES
    int setTolerance(int);      //set tolerance of sensor; returns 0, if smaller than 0
    void setShortThresh(int);   //define measuring length for short signals
    void setLongThresh(int);    //define measuring length for long signals





    //GETTER
    int isActive();
    int isCalibrated();
    int getTolerance();
    void getShortThresh(int);
    void getLongThresh(int);


};

#endif
