/* 
 * Class for integrating the arduino into the program. 
 * A usb port can be specified and the data from the serial will be read. 
 * 
 * The method "portList()" returns a String[] array with all available usb ports. 
 * with "setPort(int)"/"setPort(String portName)" the port can be selected and will be set up automatically
 * setPort() returns false, when setup has not been succcessful
 *
 * It is possible to look again for new usb ports with "updateList()"
 *
 * THE readData()-method HAS TO BE INTEGRATED INTO THE draw()-method!!
 *
 */

import processing.serial.Serial;
import java.util.Map;

interface sensorClick{
  void singleClick();
  void longClick();
}

public class ArduinoUSB implements sensorClick {
  private String[] portList;
  private String portName;
  private Serial port;
  private boolean initialized = false; //will be set to true, when port is specified
  private int tolerance = 4;
  private int checkConnectionInterval = 2000; // check of connectin is still available every x ms
  private int readDataCounter = 0;
  private int initialNumberOfPorts = 0;
  private PApplet parent;
  
  private boolean useNotifications = false;
  Notifier notifier;

  ArduinoUSB(PApplet pa) {
    parent = pa;
    this.portList = Serial.list();
    this.initialNumberOfPorts = portList.length;
    registerMethod("pre", this);
  }
  
  ArduinoUSB(PApplet pa, Notifier n) {
    parent = pa;
    this.portList = Serial.list();
    this.initialNumberOfPorts = portList.length;
    registerMethod("pre", this);
    this.notifier = n;
    useNotifications = true;
  }
  
  public void pre(){
    readPort(true);
    
  }

  public boolean setPort(String portName) { //returns true when assignment has been successful
    boolean successful = false;
    for (int i = 0; i < this.portList.length; i++) {
      if (this.portList.equals(portName)) {
        this.portName = portName;

        this.port = new Serial(parent, this.portName, 9600);
        if (this.port != null) {
          successful = true;
        }
        break;
      }
    }
    if (successful) {
      initialized = true;
    } else {
      initialized = false;
    }
    return successful;
  }

  public boolean setPort(int position) { //returns true when assignment has been successful
    boolean successful = false;
    if (position >= 0 && position < this.portList.length) {
      this.portName = this.portList[position];
      this.port = new Serial(parent, this.portName, 9600);
      if (this.port != null) {
        successful = true;
      }
    }
    if (successful) {
      initialized = true;
    } else {
      initialized = false;
    }
    return successful;
  }

  public void updateList() { //re-check for usb ports

    this.portList = Serial.list();
  }

  public String[] portList() {
    return this.portList;
  }

  public void printPortList() {
    for (int i = 0; i< this.portList.length; i++) {
      println(this.portList[i]);
    }
  }

  void readPort(boolean print) {
    readDataCounter++;
    if (readDataCounter > frameRate*checkConnectionInterval/1000) {
      readDataCounter = 0;
      updateList();
      if (portList.length < initialNumberOfPorts) {
        if(useNotifications) notifier.setNotification("Your sensor has been removed", 2000);
        initialNumberOfPorts = portList.length;
      } else if (portList.length > initialNumberOfPorts) {
        initialNumberOfPorts = portList.length;
        if (setPort(portList.length-13
        )) {
          if(useNotifications) notifier.setNotification("Sensor has been connected", 2000);
        }
      }
    }
    if (initialized) { //ensure port has been set
      while (port.available() > 0) { 
        char inChar = port.readChar(); 
        switch((int)inChar) {
        case 49: 
          singleClick();
          break;
        case 50:
          longClick();
          break;
        }
        if (print) {
          print(inChar + "");
        }
      }
    }
  }


  //____
  //interface methods:
  //_____
  void setTolerance(int t) {
    if (initialized) {
      tolerance = t;
      port.write("t" + t + "\n");
    }
  }
  int getTolerance() { //doesn't check arduino!!
    return tolerance;
  }

  void wake() {
    if (initialized) {
      port.write("1\n");
    }
  }
  void sleep() {
    if (initialized) {
      port.write("2\n");
    }
  }
  void recalibrate() {
    if (initialized) {
      port.write("3\n");
    }
  }
  void enableLongClick() {
    if (initialized) {
      port.write("4\n");
    }
  }
  void disableLongClick() {
    if (initialized) {
      port.write("5\n");
    }
  }





  //implementing the sensor interface:
  void singleClick() {

    int x = 3*mouseX/width;
    int y = 3*mouseY/height;
    interpretEyeData(x+3*y, true);
  }

  void longClick() {
    ECScreenConstants.setActiveScreen(0);
    previousDir = 4;
  }
}
