/*
 *
 *
 *
 * A custom library created by us to build a graphical user interface in processing with elements like buttons, textboxes etc.
 * 
 * 
 * Every element has to inherit from the class Control (aka Component, see below) which provides basic functions and attributes like 
 *   coordinates, size, colors etc. 
 *
 *
 * Each project that uses this library needs to create a Frame object that is the main container for the elements and initializes and calls their looks and feels.
 *
 * Use the method Frame.add(Control c); to add an element to the GUI. 
 *
 *
 * !!!
 * In addition each control needs to implement the method render(), only declared in Control. render() draws the elments looks onto it's PGraphic pg;
 * The container Frame won't call render() for it's content Controls every frame, only when the method update() has been invoked onto the object. 
 * update() sets a flag to re-render() the style next drawing time and also requests it's containers to update. This should save immense resources. 
 * !!!
 *
 *
 *
 *
 */





import java.awt.EventQueue;



public abstract class Control {

  //name can be specified manually, useful for distinguishing
  public String Name;

  //parent container and frame of object
  protected Control Parent;
  protected Frame Frame;





  //image that is of the size of the control and contains the looks
  PGraphics pg;


  //changedVisuals will be set to true when graphics of this control changed, so parents know if they need to re-render this object.
  //This will happen when update() has to be called for this Control. update() also calls update() for the parent containers
  protected boolean changedVisuals = true; 




  /*
   * Status properties
   */
  boolean Focusable = true;
  protected boolean Focused = false;        //Dont'change focused from outside
  protected boolean stickyFocus = false;    //when true than its focused state can't be overridden by other elements requesting focus, only when itself calls blur
  //textBoxes i.e. have stickyFocus, so they won't loose typing focus when the user accidently moves with the mouse over a scrollContainer


  boolean Enabled = true;
  boolean Visible = true;






  /*
   * Coordinates and size relative to container; won't be changed by flowContainer!
   */
  int X;
  int Y;

  int Height; 
  int Width;

  int MinHeight = 1;
  int MaxHeight = height;
  int MinWidth = 1;
  int MaxWidth = width;

  int MarginLeft;
  int MarginTop;
  int MarginRight;
  int MarginBottom;

  //Bound coordinates are absolute (relative to window origin), only to change by containers
  protected Bounds Bounds = new Bounds(0, 0, 0, 0);





  /*
   * Visuals
   */
  String Text;
  int FontSize;

  color BackgroundColor; 
  color ForegroundColor;
  color BorderColor;
  color HoverColor;      //first set automatically with backgroundColor, can be changed programatically
  color PressedColor;    //first set automatically with backgroundColor, can be changed programatically

  protected color VisualBackgroundColor; //usually =BackgroundColor, except when hover or pressed

  int BorderWidth;
  int Cursor;






  /*
   * Mouse Events
   */

  //Action to perfom at mouse click
  String actionPerformedMethodName;

  //previous hovered/pressed state
  protected boolean pHovered = false;
  protected boolean pPressed = false;
  protected boolean registeredMouseListener = false;










  public Control() {

    Name = "";
    Text = "";

    Height = 50;
    Width = 50;
    X = 0;
    Y = 0;
    Bounds.X0 = X;
    Bounds.Y0 = Y;
    Bounds.X = X + Width;
    Bounds.Y = Y + Width;

    BackgroundColor = 255;
    VisualBackgroundColor = 255;
    ForegroundColor = 0;
    BorderColor = 20;
    HoverColor = 200;
    PressedColor = 150;

    BorderWidth = 1;
    Cursor = 0;
    FontSize = 10;

    pg = createGraphics(Width, Height);
  }








  /*
   * DRAWING AND RENDERING
   */

  //main drawing method, must be specified for each control individually
  protected abstract void render();


  //just return the looks of this control, without drawing
  protected PImage getGraphics() {
    return pg;
  }


  //call parent to update and indicate that this control has changed its looks
  protected void update() {
    changedVisuals = true;
    Parent.update();
  }





  //specify the overall parent frame (Master GUI Container)
  protected void setParentFrame(Frame f) {
    this.Frame = f;
  }

  //request frame to focus this object
  protected void Focus() {
    Frame.requestFocus(this);
  }

  protected void Blur() {
    Frame.requestBlur(this);
  }







  /*
   * Style Setter
   */

  public void margin(int all) {
    MarginTop = all;
    MarginRight = all;
    MarginBottom = all;
    MarginLeft = all;
  }
  public void margin(int topBottom, int leftRight) {
    MarginTop = topBottom;
    MarginRight = leftRight;
    MarginBottom = topBottom;
    MarginLeft = leftRight;
  }
  public void margin(int top, int right, int bottom, int left) {
    MarginTop = top;
    MarginRight = right;
    MarginBottom = bottom;
    MarginLeft = left;
  }
  public void marginTop(int top) {
    MarginTop = top;
  }
  public void marginRight(int right) {
    MarginRight = right;
  }
  public void marginBottom(int bottom) {
    MarginBottom = bottom;
  }
  public void marginLeft(int left) {
    MarginLeft = left;
  }

  public void setText(String txt) {
    this.Text = txt;
  }
  public void setSize(int Width, int Height) {
    this.Width = Width;
    this.Height = Height;
  }






  /* 
   * EVENT METHODS
   */

  boolean addMouseListener(String methodName) {
    actionPerformedMethodName = methodName;
    registeredMouseListener = true;
    registerMethod("post", this);
    return true;
  }

  boolean removeMouseListener() {
    actionPerformedMethodName = "";
    if (registeredMouseListener) {
      unregisterMethod("post", this);
      VisualBackgroundColor = BackgroundColor; //in case button is just being hovered on or pressed
      return true;
    }
    return false;
  }



  //call the method that checks if the mouse is over control and pressing it
  public void post() {
    EventQueue.invokeLater(new mouseListener(this));
  }


  //methods that will be called when hover / pressed - events occured
  public void hover(boolean hovers, boolean isNewEvent) {
  }

  public void pressed(boolean pressed, boolean isNewEvent) {
  }




  //methods that will be called when key events occured and this Control is the currently focused element. 
  public void onKeyPress(KeyEvent e) {
  }
  public void onKeyRelease(KeyEvent e) {
  }
  public void onKeyTyped(KeyEvent e) {
  }
}





//________________________________________________________________________________________________________________________________________________________________________________
//________________________________________________________________________________________________________________________________________________________________________________


public class Container extends Control {


  public boolean AutoSize = false;

  ArrayList<Control> Content;


  Container() {
    super();
    Name ="CONTAINER";
    Width = 1;
    Height = 1;
    Content = new ArrayList<Control>();
    pg = createGraphics((int)Width, (int)Height);
  }





  /*
   * DRAWING AND RENDERING
   */


  //@override 
  public void render() {
    //if AutoSize is on, first get minimal dimensions
    if (AutoSize) {
      setAutoSize();
    }

    pg = createGraphics(Width, Height);

    pg.beginDraw();
    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }

    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);
      if (c.Visible) {

        //set bounds of child so it has absolute values for listener processing
        c.Bounds.X0 = this.Bounds.X0 + c.X;
        c.Bounds.Y0 = this.Bounds.Y0 + c.Y;
        c.Bounds.X = c.Bounds.X0 + c.Width;
        c.Bounds.Y = c.Bounds.Y0 + c.Height;

        if (c.changedVisuals) {
          c.render();
          c.changedVisuals = false;
          //println(c.Name);
        }
        //pg.image(c.CreateGraphics(), c.X, c.Y);
        pg.image(c.getGraphics(), c.X, c.Y);
      }
    }

    pg.endDraw();
  }


  public PImage getGraphics() {
    if (changedVisuals) {
      render();
      changedVisuals = false;
    }
    return pg;
  }







  //not only setting this Frame-property but also all it's children
  //@override
  void setParentFrame(Frame f) {
    this.Frame = f;
    for (int i = 0; i < Content.size(); i++) {
      Content.get(i).setParentFrame(f);
    }
  }



  public void add(Control c) {
    Content.add(c);
    c.Parent = this;
    if (c.Frame != this.Frame) {
      c.Frame = this.Frame;
    }
  }

  public void add(Control[] c) {
    for (int i = 0; i < c.length; i++) {
      Content.add(c[i]);
      if (c[i].Frame != this.Frame) {
        c[i].Frame = this.Frame;
      }
    }
  }






  //resize container so it fits its content
  public void setAutoSize() {
    int mWidth = 1;
    int mHeight = 1;
    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);
      if (c.Visible) {
        if (c.X + c.Width > mWidth) {
          mWidth = c.X + c.Width;
        }
        if (c.Y + c.Height > mHeight) {
          mHeight = c.Y + c.Height;
        }
      }
    }
    Width = mWidth;
    Height = mHeight;
  }
}



//________________________________________________________________________________________________________________________________________________________________________________
//________________________________________________________________________________________________________________________________________________________________________________



import java.lang.reflect.Method;


public class Frame extends Container {

  protected Control FocusedElement = this;







  Frame() {
    super();
    registerMethod("draw", this);
  }



  //before draw() already draw GUI, so the user may draw above it
  public void draw() {
    int t = millis();
    this.display();
    //println((millis()-t));
  }


  public void display() {
    render();
    image(pg, 0, 0);
  }




  //@override - Frame doesn't need to call its parent to update
  public void update() {
    changedVisuals = true;
  }



  //call "setParentFrame" for all children. If child is a container, it will set it's own parentFrame and call it's childrens "setParentFrame"
  public void setParentFrame() {
    for (int i = 0; i < Content.size(); i++) {
      Content.get(i).setParentFrame(this);
    }
  }



  /*
   * SHORTCUTS
   */
  class ObjectMethod {
    Method m;
    Object o;

    ObjectMethod(Method m, Object o) {
      this.m = m;
      this.o = o;
    }
  }



  HashMap<Shortcut, ObjectMethod> ShortcutMethods = new HashMap<Shortcut, ObjectMethod>();


  public boolean addShortcut(Shortcut s, String methodName, Object target) {
    Class<?> c = target.getClass();
    try {

      Method method = c.getMethod(methodName);
      ShortcutMethods.put(s, new ObjectMethod(method, target));
      //println("ADDED",ShortcutMethods.containsKey(s));
      checkShortcut(s);
    } 
    catch (NoSuchMethodException nsme) {
      die("There is no public " + methodName + "() method in the class " +
        target.getClass().getName());
    } 
    catch (Exception e) {
      die("Could not register " + methodName + " + () for " + target, e);
    }
    return false;
  }



  private boolean checkShortcut(Shortcut s) {
    ObjectMethod o =ShortcutMethods.get(s);
    if (o != null) {
      println("SHORTCUT FOUND");
      handleShortcut(o);
      return true;
    }
    return false;
  }



  private void handleShortcut(ObjectMethod om) {
    Method method = om.m;
    Object object = om.o;
    try {
      method.invoke(object, args);
    } 
    catch (Exception e) {
      // check for wrapped exception, get root exception
      Throwable t;
      /*if (e instanceof InvocationTargetException) {
       InvocationTargetException ite = (InvocationTargetException) e;
       t = ite.getCause();
       } else */       {
        t = e;
      }
      // check for RuntimeException, and allow to bubble up
      if (t instanceof RuntimeException) {
        // re-throw exception
        throw (RuntimeException) t;
      } else {
        // trap and print as usual
        printStackTrace(t);
      }
    }
  }



  /*
   * ADD CONTROLS TO CONTAINER
   */
  //when elements
  //@override

  /*void add(Control c) {
   Content.add(c);
   c.Parent = this;
   
   c.setParentFrame(this);
   }
   
   
   void add(Control[] c) {
   for (int i = 0; i < c.length; i++) {
   Content.add(c[i]);
   if (c[i].Frame != this.Frame) {
   c[i].Frame = this.Frame;
   }
   }
   }*/




  /*
   * FOCUS HANDLING
   * 
   * when other controls call requestFocus() 
   */

  public boolean requestFocus(Control c) {
    if (c.Focusable && !FocusedElement.stickyFocus) {
      FocusedElement.Focused = false;
      FocusedElement = c;
      FocusedElement.Focused = true;
      return true;
    }
    return false;
  }


  public void requestBlur(Control c) {
    if (FocusedElement == c) {
      FocusedElement.Focused = false;
      FocusedElement = this;
    }
  }







  /* 
   * KEY EVENTS
   */

  //there's only one keyListener per GUI, it hands the recognized keystrokes over to the FocusedElement
  KeyListener keyListener;

  public void addKeyEventListener() {
    registerMethod("keyEvent", this);
    keyListener = new KeyListener();
  }

  public void keyEvent(KeyEvent e) {
    keyListener.handleKeyEvent(e, FocusedElement);
  }
}

//________________________________________________________________________________________________________________________________________________________________________________
//________________________________________________________________________________________________________________________________________________________________________________








public class Button extends Control {



  public Button() {
    super();
    Name = "BUTTON";
  }






  /*
   * DRAWING AND RENDERING
   */


  //@override 
  public void render() {
    pg = createGraphics((int)Width, (int)Height);

    pg.beginDraw();
    pg.fill(VisualBackgroundColor);
    pg.stroke(BorderColor);
    pg.strokeWeight(BorderWidth);
    pg.rect(1, 1, Width-2, Height-2);

    pg.fill(ForegroundColor);
    pg.textAlign(CENTER, CENTER);
    pg.textSize(FontSize);
    pg.text(Text, Width/2, Height/2-2);
    pg.endDraw();
    changedVisuals = true;
  }







  /*
   * STYLE SETTING METHODS
   */


  //automatically set HoverColor and PressedColor

  public void setBackgroundColor(color clr) {

    BackgroundColor = clr;
    VisualBackgroundColor = clr;

    if (brightness(clr) > 40) { //darken color for HoverColor and PressedColor when color is bright enough

      float r = constrain(red(clr) - 20, 0, 255);
      float g = constrain(green(clr) - 20, 0, 255);
      float b = constrain(blue(clr) - 20, 0, 255);
      HoverColor = color(r, g, b);
      r = constrain(r - 20, 0, 255);
      g = constrain(g - 20, 0, 255);
      b = constrain(b - 20, 0, 255);
      PressedColor = color(r, g, b);
    } else {//lighten color for HoverColor and PressedColor when color too dark

      float r = constrain(red(clr) + 20, 0, 255);
      float g = constrain(green(clr) + 20, 0, 255);
      float b = constrain(blue(clr) + 20, 0, 255);
      HoverColor = color(r, g, b);
      r = constrain(r + 20, 0, 255);
      g = constrain(g + 20, 0, 255);
      b = constrain(b + 20, 0, 255);
      PressedColor = color(r, g, b);
    }
  }






  /* 
   * EVENT METHODS
   */

  //@override 
  public void hover(boolean hovers, boolean isNewEvent) {
    if (isNewEvent) {
      if (hovers) {    //enter object
        VisualBackgroundColor = HoverColor;
      } else {         //exit object
        VisualBackgroundColor = BackgroundColor;
      }
      update();
    }
  }

  //@override 
  public void pressed(boolean pressed, boolean isNewEvent) {
    if (isNewEvent) {
      update();
      if (pressed) { //pressed

        VisualBackgroundColor = PressedColor;
        if (!actionPerformedMethodName.equals("")) {
          method(actionPerformedMethodName);
        }
      } else { //released
        VisualBackgroundColor = HoverColor;
      }
    }
  }
}

//________________________________________________________________________________________________________________________________________________________________________________
//________________________________________________________________________________________________________________________________________________________________________________




public class Label extends Control {

  //no border
  public int BorderWidth = 0;
  public boolean Transparent = true;



  Label() {
    super();
    Name = "LABEL";
    VisualBackgroundColor = color(0, 0);
    pg.beginDraw();
  }








  //@override 
  public void render() {
    pg = createGraphics(Width, Height);
    pg.beginDraw();
    pg.clear();

    if (Transparent) {
      pg.noFill();
    } else {
      pg.fill(VisualBackgroundColor);
    }
    pg.noStroke();
    pg.rect(0, 0, Width, Height);

    pg.textSize(FontSize);
    pg.fill(ForegroundColor);
    pg.textAlign(CENTER, CENTER);
    pg.text(Text, Width/2, Height/2);
    pg.endDraw();
  }


  //@override 
  void setText(String s) {
    Text = trim(s);
    autoSize();
  }


  void setFontSize(int s) {
    FontSize = s;
    autoSize();
  }


  //automatic resize control, so it fits the text
  private void autoSize() {
    pg.beginDraw();
    pg.textSize(FontSize);
    Width = (int)constrain(pg.textWidth(Text), MinWidth, MaxWidth);
    Height = constrain(FontSize, MinHeight, MaxHeight);
  }
}

//________________________________________________________________________________________________________________________________________________________________________________
//________________________________________________________________________________________________________________________________________________________________________________


public class HFlowContainer extends Container {


  //@override 
  public void render() {
    //if AutoSize is on, first get minimal dimensions
    if (AutoSize) {
      setAutoSize();
    }

    pg = createGraphics(Width, Height);

    pg.beginDraw();
    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }

    int usedSpace = 0;

    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);
      if (c.Visible) {

        //set bounds of child so it has absolute values for listener processing
        c.Bounds.X0 = this.Bounds.X0 + usedSpace + c.MarginLeft;
        c.Bounds.Y0 = this.Bounds.Y0 + 0;
        c.Bounds.X = c.Bounds.X0 + c.Width;
        c.Bounds.Y = c.Bounds.Y0 + c.Height;

        if (c.changedVisuals) {
          c.render();
          c.changedVisuals = false;
          //println(c.Name);
        }
        pg.image(c.getGraphics(), usedSpace + c.MarginLeft, 0);
        usedSpace += (c.Width + c.MarginLeft + c.MarginRight);
      }
    }

    pg.endDraw();
  }
}





















public class VFlowContainer extends Container {


  //@override 
  public void render() {
    //if AutoSize is on, first get minimal dimensions
    if (AutoSize) {
      setAutoSize();
    }

    pg = createGraphics(Width, Height);
    pg.beginDraw();

    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }

    int usedSpace = 0;

    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);
      if (c.Visible) {

        //set bounds of child so it has absolute values for listener processing
        c.Bounds.X0 = this.Bounds.X0 + 0;
        c.Bounds.Y0 = this.Bounds.Y0 + usedSpace + c.MarginTop;
        c.Bounds.X = c.Bounds.X0 + c.Width;
        c.Bounds.Y = c.Bounds.Y0 + c.Height;

        if (c.changedVisuals) {
          c.render();
          c.changedVisuals = false;
          //println(c.Name);
        }
        pg.image(c.getGraphics(), 0, usedSpace + c.MarginTop);
        usedSpace += (c.Height + c.MarginTop + c.MarginBottom);
      }
    }

    pg.endDraw();
  }
}




//________________________________________________________________________________________________________________________________________________________________________________
//________________________________________________________________________________________________________________________________________________________________________________


public class HScrollContainer extends HFlowContainer {


  //Width the scrollContainer would have considering all its content (fullScrollWidth can actually be smaller that Width!)
  protected int fullScrollWidth;

  //Scroll position (only horizontal), starts at 0
  protected int scrollPositionX = 0;

  //speed at which container will be scrolled, can be set externally
  public int scrollSpeed = 10;




  //@override 
  public void render() {
    //if AutoSize is on, first get minimal dimensions
    if (AutoSize) {
      setAutoSize();
    }

    //first get total virtual width of all elements rowed up 
    fullScrollWidth = 0;
    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);

      if (c.Visible) {
        fullScrollWidth += c.MarginRight + c.Width + c.MarginLeft;
      }
    }

    pg = createGraphics(Width, Height + 5); //add some extra space for the scroll bar

    pg.beginDraw();
    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }

    int usedSpace = 0;

    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);
      if (c.Visible) {


        //basically like HFlowContainer, just subtract scrollPositionX and ensure it's not out of HScrollContainers bounds
        c.Bounds.X0 = max(this.Bounds.X0 + usedSpace + c.MarginLeft - scrollPositionX, this.Bounds.X0);
        //set bounds of child so it has absolute values for listener processing
        c.Bounds.Y0 = this.Bounds.Y0;

        //again calculate X0 (cause information has been lost when computing max() (see 3 lines above), add c.Width and ensure it's not out of HScrollContainers bounds
        c.Bounds.X = min(this.Bounds.X0 + usedSpace  + c.MarginLeft + c.Width - scrollPositionX, this.Bounds.X);

        c.Bounds.Y = c.Bounds.Y0 + c.Height;


        if (c.changedVisuals) {
          c.render();
          c.changedVisuals = false;
        }

        //subtract scrollPositionX from x-drawing-value
        pg.image(c.getGraphics(), usedSpace + c.MarginLeft - scrollPositionX, 0);
        usedSpace += (c.Width + c.MarginLeft + c.MarginRight);
      }
    }

    //Draw view-only scrollbar:
    float scrollbarWidth = (float)Width / fullScrollWidth * Width;

    if (scrollbarWidth < Width) { //don't display scroll-bar when there's nothign to scroll

      float scrollBarX = constrain(scrollPositionX * (float)(Width - scrollbarWidth) / (float)(fullScrollWidth - Width), 1, Width - scrollbarWidth - 2);
      pg.fill(150);
      pg.noStroke();
      pg.rect(scrollBarX, Height + 1, scrollbarWidth, 3, 15);
    }

    pg.endDraw();
  }





  public void addToScrollPositionX(float add) {
    scrollPositionX = (int)constrain(scrollPositionX + (add * scrollSpeed), 0, max(0, fullScrollWidth - Width));
    update();
  }
  public void setScrollPositionX(float i) {
    scrollPositionX = (int)constrain(i, 0, fullScrollWidth - Width);
    update();
  }




  /*
   * MOUSE EVENTS
   */
   
  //@override 
  public void hover(boolean hovers, boolean isNewEvent) {
    if (isNewEvent) {
      if (hovers) {
        this.Focus();
        update();
      } else {
        this.Blur();
      }
      update();
    }
  }
}










public class VScrollContainer extends VFlowContainer {

  //Width the scrollContainer would have considering all its content (fullScrollHeight can actually be smaller that Height!)
  protected int fullScrollHeight;

  //Scroll position (only vertical), starts at 0
  protected int scrollPositionY = 0;

  //speed at which container will be scrolled, can be set externally
  public int scrollSpeed = 10;


  //@override 
  public void render() {
    //if AutoSize is on, first get minimal dimensions
    if (AutoSize) {
      setAutoSize();
    }

    //first get total virtual width of all elements rowed up 
    fullScrollHeight = 0;
    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);

      if (c.Visible) {
        fullScrollHeight += c.MarginTop + c.Height + c.MarginBottom;
      }
    }

    pg = createGraphics(Width + 5, Height); //add some extra space for the scroll bar

    pg.beginDraw();
    if (VisualBackgroundColor != 0) {
      pg.background(VisualBackgroundColor);
    }

    int usedSpace = 0;

    for (int i = 0; i < Content.size(); i++) {
      Control c = Content.get(i);
      if (c.Visible) {

        //set bounds of child so it has absolute values for listener processing
        c.Bounds.X0 = this.Bounds.X0 + 0;
        //basically like VFlowContainer, just subtract scrollPositionY and ensure it's not out of VScrollContainers bounds
        c.Bounds.Y0 = max(this.Bounds.Y0 + usedSpace + c.MarginTop - scrollPositionY, this.Bounds.Y0);

        c.Bounds.X = c.Bounds.X0 + c.Width;
        //again calculate Y0 (cause information has been lost when computing max() (see 3 lines above), add c.Height and ensure it's not out of VScrollContainers bounds
        c.Bounds.Y = min(this.Bounds.Y0 + usedSpace  + c.MarginTop + c.Height - scrollPositionY, this.Bounds.Y);


        if (c.changedVisuals) {
          c.render();
          c.changedVisuals = false;
        }

        //subtract scrollPositionX from x-drawing-value
        pg.image(c.getGraphics(), c.MarginLeft, usedSpace + c.MarginTop - scrollPositionY);
        usedSpace += (c.Height + c.MarginTop + c.MarginBottom);
      }
    }

    //Draw view-only scrollbar:
    float scrollbarHeight = (float)Height / fullScrollHeight * Height;

    if (scrollbarHeight < Height) { //don't display scroll-bar when there's nothign to scroll

      float scrollBarY = constrain(scrollPositionY * (float)(Height - scrollbarHeight) / (float)(fullScrollHeight - Height), 1, Height - scrollbarHeight - 2);
      pg.fill(200);
      pg.noStroke();
      pg.rect(Width + 1, scrollBarY, 3, scrollbarHeight, 15);
    }

    pg.endDraw();
  }






  public void addToScrollPositionY(float add) {
    scrollPositionY = (int) constrain(scrollPositionY + (add * scrollSpeed), 0, max(0, fullScrollHeight - Height));
    update();
  }
  public void setScrollPositionY(float i) {
    scrollPositionY = (int) constrain(i, 0, max(0, fullScrollHeight - Height));
    update();
  }


  /*
   * MOUSE EVENTS
   */
   
  //@override 
  public void hover(boolean hovers, boolean isNewEvent) {
    if (isNewEvent) {
      if (hovers) {
        this.Focus();
        update();
      } else {
        this.Blur();
      }
      update();
    }
  }
}



//________________________________________________________________________________________________________________________________________________________________________________




public class mouseListener implements Runnable {
  Control Ctrl;

  mouseListener(Control btn) {
    this.Ctrl = btn;
  }
  final boolean PRESSED = true;
  final boolean NotPRESSED = false;
  final boolean NEWEVENT = true;
  final boolean NoNEWEVENT = false;

  //event listener checker
  public void run() {
    if (Ctrl.Visible && Ctrl.Enabled) {

      if (mouseX > Ctrl.Bounds.X0 && mouseX < Ctrl.Bounds.X && mouseY > Ctrl.Bounds.Y0 && mouseY  < Ctrl.Bounds.Y) {

        if (mousePressed) { //when mouse press start

          if (!Ctrl.pPressed) {
            Ctrl.pPressed = true;
            Ctrl.pressed(PRESSED, NEWEVENT);
          } else {
            Ctrl.pressed(PRESSED, NoNEWEVENT);
          }
        } else if (!mousePressed && Ctrl.pPressed) { //when mouse release

          Ctrl.pressed(NotPRESSED, NEWEVENT);
          Ctrl.pPressed = false;
        } else {
          if (!Ctrl.pHovered) {
            Ctrl.pHovered = true;
            Ctrl.hover(true, NEWEVENT);
          }
        }
      } else {
        if (Ctrl.pHovered) {
          Ctrl.pHovered = false;
          Ctrl.hover(false, NEWEVENT);
        }
      }
    }
  }
}












public class KeyListener {

  final int CONTROL = 17;
  final int SHIFT = 16;
  final int ALT = 18;
  final int PRESS = 0;
  final int TYPED = 1;
  final int RELEASE = 2;

  public boolean control = false;   //true when control is pressed
  public boolean shift = false;     //true when shift is pressed
  public boolean alt = false;       //true when alt is pressed

  public int previousKeyCode;
  public char previousKey;

  public int currentKeyCode;
  public char currentKey;

  ArrayList<Long> pressedKeys = new ArrayList<Long>(6);

  public KeyListener() {
  }

  public void handleKeyEvent(KeyEvent event, Control c) {

    previousKeyCode = currentKeyCode;
    previousKey = currentKey;
    int lastAction = 0;

    switch(event.getAction()) {
    case KeyEvent.PRESS: 
      c.onKeyPress(event);
      lastAction = PRESS;
      break;
    case KeyEvent.TYPE:
      c.onKeyTyped(event);
      lastAction = TYPED;
      break;
    case KeyEvent.RELEASE:
      c.onKeyRelease(event);
      lastAction = RELEASE;
      break;
    }

    if (key == CODED) {
      switch(keyCode) {
      case CONTROL:
        if (lastAction == PRESS) {
          control = true;
        } else if (lastAction == RELEASE) {
          control = false;
        }
        break;
      case SHIFT:
        if (lastAction == PRESS) {
          shift = true;
        } else if (lastAction == RELEASE) {
          shift = false;
        }
        break;
      case ALT:
        if (lastAction == PRESS) {
          alt = true;
        } else if (lastAction == RELEASE) {
          alt = false;
        }
        break;
      }
    } 



    //handle shortcuts
    if (lastAction == TYPED) {
    } else if (lastAction == PRESS) {

      Shortcut current = new Shortcut((char)keyCode, control, shift, alt);
      //println("checkshort", (char)keyCode, control, shift, alt, current.equals(sc), (current.hashCode() == sc.hashCode()));
      //Frame.checkShortcut(current);
    }
  }
}



//________________________________________________________________________________________________________________________________________________________________________________


public class Bounds {
  int X0;
  int X;
  int Y;
  int Y0;

  Bounds(int X0, int Y0, int X, int Y) {
    this.X0 = X0;
    this.Y0 = Y0;
    this.X = X;
    this.Y = Y;
  }


  void print() {
    println(X0, Y0, X, Y);
  }
}

/*
 * Shortcuts can be created with Shortcut(true, true, false, 'A') for Ctrl-Shift-A
 * or like Shortcut(A, CONTROL, SHFIT);
 *
 *
 */



public class Shortcut {

  public boolean Control = false;
  public boolean Shift = false;
  public boolean Alt = false;
  public char Key;



  Shortcut(char Key, boolean Control, boolean Shift, boolean Alt) {
    this.Control = Control;
    this.Shift = Shift;
    this.Alt = Alt;
    this.Key = Key;
  }


  Shortcut(char Key, int...Modifiers) {
    this.Key = Key;
    for (int modifier : Modifiers) {
      switch(modifier) {
      case 17:
        this.Control = true;
        break;
      case 16:
        this.Shift = true; 
        break;
      case 18:
        this.Alt = true;
        break;
      }
    }
  }
  
  @Override
    public boolean equals(Object other) {
    if (this == other) return true;
    if(getClass() != other.getClass()) return false;
    
    Shortcut shortcut = (Shortcut) other;
    
    if (Control == shortcut.Control && Shift == shortcut.Shift && Alt == shortcut.Alt && Key == shortcut.Key) {
      return true;
    }
    return false;
  }

  @Override
    public int hashCode() {
    int result = (int)Key;
    result = result * 100 + (Control ? 1 : 0);
    result = result * 10 + (Shift ? 1 : 0);
    result = result * 10 + (Alt ? 1 : 0);
    return result;
  }
}
